;;; qpdfedit.el --- edit .pdf by qpdf .qdf

;; Copyright 2019, 2020 Kevin Ryde

;; Author: Kevin Ryde <user42@zip.com.au>
;; Version: 0
;; Keywords: data, pdf
;; URL: http://user42.tuxfamily.org/qpdfedit/index.html

;; qpdfedit.el is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; qpdfedit.el is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
;; Public License for more details.
;;
;; You can get a copy of the GNU General Public License online at
;; <http://www.gnu.org/licenses/>.

;;; Commentary:

;; This is a buffer-file-format mechanism to decode .pdf using the qpdf
;; program and its "qdf" format for viewing and editing.  Editing is in
;; Postscript mode (`ps-mode') and re-writing is by qpdf (with its fix-qdf
;; to adjust object indexes automatically).
;;
;; PDF is a poor format for actual editing, and can be made hopeless or
;; nearly so depending on generator tools and how much is done with bitmap
;; images.  In all cases it tends to be fragile so be sure to keep backups.
;; See `qpdfedit' help string for more.

;;; Install:

;; To edit .pdf this way on visiting, put qpdfedit.el in one of your
;; `load-path' directories and the following in your .emacs
;;
;;     (autoload 'qpdfedit "qpdfedit" nil t)
;;     (modify-coding-system-alist 'file "\\.pdf\\'" 'raw-text-unix)
;;     (add-to-list 'auto-mode-alist '("\\.pdf\\'" . qpdfedit))
;;
;; There's an autoload cookie for `qpdfedit' below, but not the .pdf
;; filename coding and mode.  Emacs has a builtin `doc-view-mode' for
;; visiting .pdf and overriding that is left to user preference.

;;; History:

;; Version 1 - the first version

;;; Emacsen:

;; Designed for Emacs 21 and up, works in XEmacs 21.
;; Might work in Emacs 20 if you have poe.el for `make-temp-file'.

;;; Code:


;;-----------------------------------------------------------------------------
;; compatibility

;; `make-temp-file' new in emacs21, not in xemacs21
(cond ((or (eval-when-compile (fboundp 'make-temp-file))
           (fboundp 'make-temp-file))
       ;; emacs21 up, noticed at compile time or run time
       (eval-and-compile
         (defalias 'qpdfedit--make-temp-file 'make-temp-file)))

      ((locate-library "mm-util") ;; from gnus
       ;; xemacs21
       (eval-and-compile
         (autoload 'mm-make-temp-file "mm-util")
         (defalias 'qpdfedit--make-temp-file 'mm-make-temp-file)))

      ((locate-library "poe") ;; from APEL
       ;; emacs20 with poe.el add-on
       (require 'poe)
       (eval-and-compile
         (defalias 'qpdfedit--make-temp-file 'make-temp-file)))

      (t
       ;; umm, dunno, hope the user can define it
       (message "qpdfedit.el: don't know where to get `make-temp-file'")
       (defalias 'qpdfedit--make-temp-file 'make-temp-file)))

;; `set-buffer-multibyte' not in xemacs21
(eval-and-compile
  (defalias 'qpdfedit--set-buffer-multibyte
    (if (eval-when-compile (fboundp 'set-buffer-multibyte))
        'set-buffer-multibyte  ;; emacs
      'identity)))             ;; not applicable in xemacs21


;;-----------------------------------------------------------------------------
;; generic

(eval-when-compile
  (defmacro qpdfedit--with-errorfile (&rest body)
    "An internal part of qpdfedit.el.
This macro doesn't exist when running byte compiled.

Create a temporary file for use by the BODY forms.
Variable `errorfile' is the filename.
An `unwind-protect' ensures the file is removed no matter what
BODY does."
    ;; (declare (debug t))  ;; emacs22,xemacs21, or 'cl
    `(let ((errorfile (qpdfedit--make-temp-file "qpdfedit-el-")))
       (unwind-protect
           (progn ,@body)
         (delete-file errorfile)))))

;; quieten byte compiler pre-emacs23
(defvar write-region-post-annotation-function)

(eval-when-compile
  (defmacro qpdfedit--without-post-kill (&rest body)
    "An internal part of qpdfedit.el.
This macro doesn't exist when running byte compiled.

Evaluate BODY without post-annotation kill-buffer.
If `write-region-post-annotation-function' is set buffer-local to
`kill-buffer' then set it to nil for BODY, and restore by an
`unwind-protect' afterwards.

This is a workaround for a bug in Emacs 23.1 where
`write-region-post-annotation-function' is set to `kill-buffer',
meaning any writes done by an encode function kill the buffer
that the encode is supposed to be operating on, usually making it
go on to mangle the contents of an unrelated buffer."

    ;; (declare (debug t))  ;; emacs22,xemacs21, or 'cl
    `(let* ((qpdfedit--without-post-kill--bad
             (and (local-variable-p 'write-region-post-annotation-function
                                    (current-buffer))
                  (eq write-region-post-annotation-function
                      'kill-buffer)))
            (qpdfedit--without-post-kill--buffer (current-buffer)))
       (unwind-protect
           (progn
             (if qpdfedit--without-post-kill--bad
                 (setq write-region-post-annotation-function nil))
             ;; (message "buf  %S" qpdfedit--without-post-kill--buffer)
             ;; (message " bad  %S" qpdfedit--without-post-kill--bad)
             ;; (message " now  %S" write-region-post-annotation-function)
             ,@body)
         (and qpdfedit--without-post-kill--bad
              (buffer-live-p qpdfedit--without-post-kill--buffer)
              (with-current-buffer qpdfedit--without-post-kill--buffer
                (set (make-local-variable 'write-region-post-annotation-function)
                     'kill-buffer)))))))

(defun qpdfedit-switch-to-buffer-other-window (buffer)
  "An internal part of qpdfedit.el.
Switch to display BUFFER in another window.
If it wasn't already in a window then the window is shrunk with
`shrink-window-if-larger-than-buffer'."
  (let ((existing-window (get-buffer-window buffer)))
    (condition-case nil
        ;; emacs two args
        (switch-to-buffer-other-window buffer t) ;; no-record
      (error
       ;; xemacs one arg
       (switch-to-buffer-other-window buffer)))
    (if (not existing-window)
        (shrink-window-if-larger-than-buffer
         (get-buffer-window buffer)))))


;;-----------------------------------------------------------------------------
;; format

;; No automatic detection here.
;;
(add-to-list 'format-alist '(qpdfedit
                             "PDF converted to QDF for editing."
                             nil ;; no automatic decode
                             qpdfedit-decode
                             qpdfedit-encode
                             t     ;; encode modifies the region
                             nil)) ;; write removes from buffer-file-formats

(defun qpdfedit-decode (beg end)
  ;; checkdoc-params: (beg end)
  "Run qpdfedit on raw .pdf bytes in the current buffer.
This function is for use from `format-alist'.

The buffer should be unibyte as per a `raw-text-unix' read.  The
bytes are put through \"qpdf --qdf\" to get mostly text and the
buffer is switched to multibyte.  An error is thrown if qpdf
can't be run or the buffer contents are invalid."

  (save-excursion
    (save-restriction
      (narrow-to-region beg end)
      ;; fairly sure qpdfedit spits out ascii-only, but go `undecided'
      (qpdfedit-run "qpdf" '("--qdf") 'raw-text-unix 'raw-text-unix nil)
      (qpdfedit--set-buffer-multibyte t)
      (point-max))))

(defun qpdfedit-encode (beg end buffer)
  ;; checkdoc-params: (beg end buffer)
  "Run bdftopcf on bdf text in the current buffer.
This function is for use from `format-alist'.

The buffer text is put through bdftopcf to produce pcf bytes,
which replace the text and the buffer is switched to unibyte.
An error is thrown if bdftopcf can't be run or the buffer
contents are invalid."

  (qpdfedit--without-post-kill
   (save-excursion
     (save-restriction
       (narrow-to-region beg end)
       ;; dunno what input coding qpdf expects, maybe means to be little
       ;; bits of binary only
       (qpdfedit-run "qpdf-with-fix" '()
                     ;; "qpdf" '("--stream-data=compress"
                     ;;          "--compress-streams=y"
                     ;;          "--ignore-xref-streams")
                     'raw-text-unix 'raw-text-unix buffer)
       (point-max)))))

(defun qpdfedit-run (command args write-coding read-coding originating-buffer)
  ;; checkdoc-params: (args write-coding read-coding originating-buffer)
  "An internal part of qpdfedit.el.
Run COMMAND on the current buffer.  ARGS is a list of strings.
Program output replaces the buffer contents.

ORIGINATING-BUFFER is the source buffer for when running qpdf.
This might be different from the current buffer.  It's recorded
as the place to go for errors."

  (qpdfedit--with-errorfile
   (let ((input-buffer (current-buffer))
         (infilename   (make-temp-file "qpdfedit-in-"))
         (outfilename  (make-temp-file "qpdfedit-out-"))
         status)
     (setq args (cons infilename (cons outfilename args)))
     (with-temp-file infilename
       (with-current-buffer input-buffer
         (write-region (point-min) (point-max) infilename))
       ;; nil     ;; append
       ;; nil     ;; visit
       ;; nil     ;; lockname
       ;; 'excl)) ;; new file, or error out

       (with-temp-file outfilename
         (with-temp-buffer
           (let ((temp-buffer (current-buffer)))
             (with-current-buffer input-buffer
               (let ((coding-system-for-write write-coding)
                     (coding-system-for-read  read-coding)
                     (default-directory       "/")
                     (process-connection-type nil)) ;; pipe
                 (setq status (apply
                               'call-process-region
                               (point-min) (point-max)
                               command
                               nil               ;; don't delete old
                               (list temp-buffer ;; stdout to temp buffer
                                     errorfile)  ;; stderr to file
                               nil               ;; no redisplay
                               args)))

               ;; replace input-buffer with new if `call-process-region' didn't
               ;; error out
               (insert-file-contents outfilename nil nil nil t))))))

     (if (equal 0 status)
         (let ((buffer (get-buffer "*qpdfedit-errors*")))
           (when buffer
             (delete-windows-on buffer)
             (kill-buffer buffer)))

       (save-selected-window
         (with-current-buffer (get-buffer-create "*qpdfedit-errors*")
           (setq buffer-read-only nil)
           (fundamental-mode)
           (erase-buffer)

           ;; emacs21 ignores the first two lines of a compilation-mode
           ;; buffer, so add in dummies
           (if (numberp status)
               (insert (format "%s exit %s\n\n" command status))
             (insert (format "%s %s\n\n" command status)))

           ;; matched by `qpdfedit--filename' compile pattern above
           (insert "### qpdfedit.el input:\n")

           (insert-file-contents errorfile)

           (qpdfedit-switch-to-buffer-other-window (current-buffer))
           (goto-char (point-min))
           (if originating-buffer
               (compilation-mode))
           (setq qpdfedit-originating-buffer originating-buffer)
           (error "%s error, see *qpdfedit-errors* buffer" command))))

     ;; cleanup temp files
     (ignore-errors (delete-file infilename))
     (ignore-errors (delete-file outfilename)))))

;;-----------------------------------------------------------------------------
;; apply format

;;;###autoload
(defun qpdfedit ()
  "Decode a pdf to qdf text format using the qpdf program.
This function is designed for use from `auto-mode-alist'.
The buffer should be raw pdf bytes (`raw-text-unix' unibyte).

The bytes are decoded by the \"qpdf\" program to its \"qdf\" text
nearly human readable form.  Emacs `format-alist' mechanism is
used so that the buffer contents are re-coded to pdf on saving
changes.  The buffer is put into Postscript `ps-mode' when that
mode exists, otherwise `fundamental-mode'.

Viewing qdf can be of interest to see generator info or style.
Re-writing can apply minor edits like spelling, capitalization,
or delete bits of text, though PDF is generally a poor format for
editing.  Making changes might depend on the way fonts are done,
and changing even just capitalization may make a letter disappear
or upset the font displayed.

PDF often has words are drawn at fixed positions on the
page (rather than relative to previous word) so inserting extra
text may overwrite what follows.  Images usually can't be
changed.  In all cases editing tends to be hairy, so make a
backup first.  PDF has various object sizes and indexes.  qpdf's
\"qdf-fix\" program is used to correct those as necessary on
saving.

If `auto-compression-mode' is enabled then .pdf.gz files can be
read and written too.  .pdf.gz is usually not needed since pdf
has its own \"stream\" compression.

For saving, note that `tar-mode' (as of Emacs 23) doesn't follow
`buffer-file-format' so writing a .pdf inside a .tar gives just
the qdf text.  This afflicts all file format things.  So don't do
that.  `archive-mode' saving is ok.

The qpdf program home page is
  URL `http://github.com/qpdf/'
The qpdfedit.el home page is
  URL `http://user42.tuxfamily.org/qpdfedit/index.html'"

  (unless (memq 'qpdfedit buffer-file-format)
    (let ((inhibit-read-only t))
      (format-decode-buffer 'qpdfedit)))

  (if (fboundp 'ps-mode) ;; if a hypothetical bdf-mode exists
      (ps-mode)
    (fundamental-mode)))


;;-----------------------------------------------------------------------------

;; LocalWords: qpdf qpdfedit unibyte gz builtin ok el

(provide 'qpdfedit)

;;; qpdfedit.el ends here
