;; Copyright 2009, 2010, 2011, 2012, 2014, 2016 Kevin Ryde

;; This file is part of miscbits-el.
;;
;; miscbits-el is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; miscbits-el is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
;; Public License for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with miscbits-el; see the file COPYING.  Failing that, go to
;; <http://www.gnu.org/licenses/>.

(add-to-list 'load-path (expand-file-name ".."))

(defconst my-sqlite2-available-p
  ;; "sqlite -version" exits with code 1, but stdin from /dev/null exits 0
  (condition-case nil
      (equal 0 (call-process "sqlite"
                             nil   ;; stdin
                             nil   ;; stdout
                             nil)) ;; display
    (error nil))
  "Non-nil if the \"sqlite\" program is available.")

(defconst my-sqlite3-available-p
  (condition-case nil
      (equal 0 (call-process "sqlite3"
                             nil  ;; stdin
                             nil  ;; stdout
                             nil  ;; display
                             "-version"))
    (error nil))
  "Non-nil if the \"sqlite3\" program is available.")

(defun my-run-test ()
  (require 'miscbits-el-loaddefs)
  (let ((count 0))

    ;;-------------------------------------------------------------------------

    (unless (fboundp 'sql-mode)
      (error "oops, `sql-mode' not available"))
    (setq count (1+ count))

    ;;-------------------------------------------------------------------------
    ;; load

    (require 'sqlite-dump)

    ;; no leftovers in emacs21 byte compiling defadvice
    (when (and (boundp 'byte-compile-unresolved-functions)
               byte-compile-unresolved-functions)
      (error "byte-compile-unresolved-functions after load: %S"
             byte-compile-unresolved-functions))
    (setq count (1+ count))

    (require 'compile)

    ;; no leftovers in emacs21 byte compiling defadvice
    (when (and (boundp 'byte-compile-unresolved-functions)
               byte-compile-unresolved-functions)
      (error "byte-compile-unresolved-functions after load 'compile: %S"
             byte-compile-unresolved-functions))
    (setq count (1+ count))


    ;;-------------------------------------------------------------------------
    ;; compile eval-after-load working

    (or (and (boundp 'compilation-error-regexp-systems-list)
             ;; xemacs21
             (message "compilation-error-regexp-systems-list (xemacs21)")
             (assq 'sqlite-dump--sqlite3 compilation-error-regexp-alist-alist))

        (and (boundp 'compilation-error-regexp-alist-alist)
             ;; emacs22
             (message "compilation-error-regexp-alist-alist (emacs22)")
             (memq 'sqlite-dump--sqlite3 compilation-error-regexp-alist))

        ;; emacs21
        (message "compilation-error-regexp-alist (emacs21)")
        (let (found)
          (mapcar (lambda (elem)
                    (setq elem (car-safe elem))
                    (and (stringp elem)
                         (string-match "Error: near line" elem)
                         (setq found t)))
                  compilation-error-regexp-alist)
          found)

        (error "sqlite-dump compile regexp not applied"))
    (setq count (1+ count))


    ;;-------------------------------------------------------------------------
    ;; sqlite-dump--without-post-kill -- macro

    (let (got
          (want nil))
      (sqlite-dump--without-post-kill
       (setq got (and (boundp 'write-region-post-annotation-function)
                      write-region-post-annotation-function)))
      (unless (equal got want)
        (error "without-post-kill: unset got %S want %S"
               got want)))
    (setq count (1+ count))

    (with-temp-buffer
      (set (make-local-variable 'write-region-post-annotation-function)
           'kill-buffer)
      (let (got
            (want nil))
        (sqlite-dump--without-post-kill
         (setq got (and (boundp 'write-region-post-annotation-function)
                        write-region-post-annotation-function)))
        (unless (equal got want)
          (error "without-post-kill: kill got %S want %S"
                 got want))
        (unless (equal write-region-post-annotation-function 'kill-buffer)
          (error "without-post-kill: kill post not restored"))))
    (setq count (1+ count))

    (let ((func (lambda (&rest args) 'noop)))
      (with-temp-buffer
        (set (make-local-variable 'write-region-post-annotation-function) func)
        (let (got
              (want func))
          (sqlite-dump--without-post-kill
           (setq got (and (boundp 'write-region-post-annotation-function)
                          write-region-post-annotation-function)))
          (unless (equal got want)
            (error "without-post-kill: another got %S want %S"
                   got want)))
        (let ((got  write-region-post-annotation-function)
              (want func))
          (unless (equal got want)
            (error "without-post-kill: another out got %S want %S"
                   got want)))
        (setq count (1+ count))))

    ;;-------------------------------------------------------------------------
    ;; sqlite-dump--with-tempfile -- macro

    (if (not (fboundp 'sqlite-dump--with-tempfile))
        (message "skip, no sqlite-dump--with-tempfile macro after byte compile")

      (sqlite-dump--with-tempfile
       (unless (boundp 'tempfile)
         (error "sqlite-dump--with-tempfile doesn't bind `tempfile'"))
       (message "t-sqlite-dump.el: tempfile is %S" tempfile))
      (setq count (1+ count)))

    ;;-------------------------------------------------------------------------
    ;; sqlite-dump--with-tempdbfile -- macro

    (if (not (fboundp 'sqlite-dump--with-tempdbfile))
        (message "skip, no sqlite-dump--with-tempfile macro after byte compile")

      (sqlite-dump--with-tempdbfile
       (unless (boundp 'tempdbfile)
         (error "sqlite-dump--with-tempdbfile doesn't bind `tempdbfile'"))
       (message "t-sqlite-dump.el: tempdbfile is %S" tempdbfile)
       (unless (file-name-absolute-p tempdbfile)
         (error "sqlite-dump--with-tempdbfile `tempdbfile' not absolute")))
      (setq count (1+ count))

      ;; removing db file
      (sqlite-dump--with-tempdbfile
       (if (file-exists-p tempdbfile)
           (error "sqlite-dump--with-tempdbfile `tempdbfile' exists already"))
       (write-region "hello" nil tempdbfile))
      (setq count (1+ count))

      ;; removing db and journal files
      (sqlite-dump--with-tempdbfile
       (let ((journalfile (concat tempdbfile "-journal")))
         (if (file-exists-p tempdbfile)
             (error "sqlite-dump--with-tempdbfile `tempdbfile' exists already"))
         (if (file-exists-p journalfile)
             (error "sqlite-dump--with-tempdbfile `tempdbfile' journal exists"))
         (write-region "hello" nil tempdbfile)
         (write-region "world" nil journalfile)))
      (setq count (1+ count)))


    ;;-------------------------------------------------------------------------
    (if (not my-sqlite2-available-p)
        (message "  skip, sqlite2 program not available")

      ;; decode working from a non-existent directory
      (let ((buffer (get-buffer-create "*t-sqlite-dump test*")))
        (switch-to-buffer buffer)
        (if (fboundp 'set-buffer-multibyte) ;; not in xemacs21
            (set-buffer-multibyte nil))
        (insert-file-contents-literally "t-sqlite-dump.sq2")
        (setq default-directory "/no/such/dir/exists/")
        (message "t-sqlite-dump.el: decode t-sqlite-dump.sq2")
        (sqlite-dump)
        (kill-buffer buffer)
        (setq count (1+ count)))

      ;;----------------------------------------------------------------------
      ;; saving

      (condition-case nil
          (delete-file "t-sqlite-dump--temp.sq2") (error nil))
      (copy-file "t-sqlite-dump.sq2" "t-sqlite-dump--temp.sq2")

      (unwind-protect
          (progn
            (message "t-sqlite-dump.el: read t-sqlite-dump--temp.sq2")
            (find-file "t-sqlite-dump--temp.sq2")
            (sqlite-dump)

            (message "t-sqlite-dump.el: write t-sqlite-dump--temp.sq2")
            (set-buffer-modified-p t)
            (save-buffer)
            (kill-buffer nil)

            (message "t-sqlite-dump.el: re-read t-sqlite-dump--temp.sq2")
            (find-file "t-sqlite-dump--temp.sq2")
            (sqlite-dump)
            (kill-buffer nil)

            (setq count (1+ count)))

        (condition-case nil
            (delete-file "t-sqlite-dump--temp.sq2") (error nil))
        (condition-case nil
            (delete-file "t-sqlite-dump--temp.sq2~") (error nil))))


    ;;-------------------------------------------------------------------------
    (if (not my-sqlite3-available-p)
        (message "  skip, sqlite3 program not available")

      ;; decode working from a non-existent directory
      (let ((buffer (get-buffer-create "*t-sqlite-dump test*")))
        (switch-to-buffer buffer)
        (if (fboundp 'set-buffer-multibyte) ;; not in xemacs21
            (set-buffer-multibyte nil))
        (insert-file-contents-literally "t-sqlite-dump.sq3")
        (setq default-directory "/no/such/dir/exists/")
        (message "t-sqlite-dump.el: decode t-sqlite-dump.sq3")
        (sqlite-dump)
        (kill-buffer buffer)
        (setq count (1+ count)))

      ;;----------------------------------------------------------------------
      ;; saving

      (condition-case nil
          (delete-file "t-sqlite-dump--temp.sq3") (error nil))
      (copy-file "t-sqlite-dump.sq3" "t-sqlite-dump--temp.sq3")

      (unwind-protect
          (progn
            (message "t-sqlite-dump.el: read t-sqlite-dump--temp.sq3")
            (find-file "t-sqlite-dump--temp.sq3")
            (sqlite-dump)

            (message "t-sqlite-dump.el: write t-sqlite-dump--temp.sq3")
            (set-buffer-modified-p t)
            (save-buffer)
            (kill-buffer nil)

            (message "t-sqlite-dump.el: re-read t-sqlite-dump--temp.sq3")
            (find-file "t-sqlite-dump--temp.sq3")
            (sqlite-dump)
            (kill-buffer nil)

            (setq count (1+ count)))

        (condition-case nil
            (delete-file "t-sqlite-dump--temp.sq3") (error nil))
        (condition-case nil
            (delete-file "t-sqlite-dump--temp.sq3~") (error nil))))

    ;;-------------------------------------------------------------------------
    ;; unload-feature

    (add-to-list 'load-path (expand-file-name "."))
    (require 'my-unload)

    (if (not my-unload-feature-function-p)
        (message "skip, no unload-feature function")

      (let ((feature 'sqlite-dump))
        (unless (my-advice-by-name feature)
          (error "oops, didn't find any advice"))

        (unload-feature feature t)

        (let ((lst (my-advice-by-name feature)))
          (when lst
            (error "unload-feature leftover advice: %S" lst)))
        (setq count (1+ count))))


    ;;-----------------------------------------------------------------------
    ;; unload-feature -- with advice unloaded first

    (let ((feature 'sqlite-dump))
      (if (featurep feature)
          (unload-feature feature t))
      (require feature)
      ;; if 'advice can be unloaded (no dependency expressed on it), then
      ;; see FEATURE unloads successfully too
      (when (condition-case nil
                (progn (unload-feature 'advice) t)
              (error nil))
        (unload-feature feature t)))
    (setq count (1+ count))


    ;;-----------------------------------------------------------------------
    ;; compilation-find-file -- after unload-feature

    (compilation-find-file nil "t-sqlite-dump.el" nil)
    (setq count (1+ count))


    ;;-----------------------------------------------------------------------
    ;; next-error -- after unload-feature

    ;; something fishy in emacs22,23 "Moved past last error" trying to do a
    ;; programmatic next-error like this ...

    (require 'sqlite-dump)
    (find-file "t-sqlite-dump--compilation.txt")
    (compilation-mode)

    ;; (if (featurep 'sqlite-dump)
    ;;     (unload-feature 'sqlite-dump t))
    ;; (message "sqlite-dump-originating-buffer boundp: %s"
    ;;          (boundp 'sqlite-dump-originating-buffer))

    (switch-to-buffer "t-sqlite-dump--compilation.txt")
    (goto-char (point-min))

    ;; (goto-char (point-max))
    ;; (forward-line -1)
    ;; (message "buffer-name %S" (buffer-name))
    ;; (message "line %S" (thing-at-point 'line))
    ;; (message "line num %S" (count-lines (point-min) (point)))
    ;; (message "mode %S" major-mode)
    ;; (compile-goto-error)
    ;; (message "buffer %S" (current-buffer))

    (let ((skip nil))
      (condition-case err
          (next-error)
        (error
         (setq skip t)
         (message "Oops, next-error: %S\ndummy out this test"
                  err)))

      ;; something fishy in emacs24 -batch
      (setq skip t)

      (unless skip
        (let ((got  (buffer-name))
              (want "t-sqlite-dump.el"))
          (unless (equal got want)
            (error "next-error buffer got %S want %S" got want)))
        (let ((got  (point))
              (want (point-min)))
          (unless (equal got want)
            (error "next-error point got %S want %S" got want)))
        (setq count (1+ count))))


    ;;-------------------------------------------------------------------------

    (message "  %d tests" count)
    count))
